\insert 'functions.oz'

declare SudokuLists IASudominoku IA

% Función para sacar las listas en donde sus números deben ser todos diferentes
fun {SudokuLists T}
   Lists   = {NewCell nil}
   Rows    = {NewCell nil}
   Columns = {NewCell nil}
   Squares = {NewCell nil}
in
   for I in 1..9 do
      Rows    := nil
      Columns := nil
      Squares := nil
      for J in 1..9 do
	 Rows    := {List.append @Rows [{List.nth T 9*(I-1) + J}]}    % Fila
	 Columns := {List.append @Columns [{List.nth T 9*(J-1) + I}]} % Columna 
	 Squares := {List.append @Squares [{List.nth T ((I-1) div 3 * 27) + ((I-1) mod 3 * 3) + ((J-1) div 3 * 9) + ((J-1) mod 3 + 1)}]} % Cuadro
      end
      Lists := {List.append @Lists [@Rows @Columns @Squares]}
   end
   @Lists
end

% Función para acar las posibles posiciones donde se puede poner una ficha (0=No 1:Sí)
fun {PossiblePositions T}
   if T == nil then nil else
      if {IsDet T.1} then 0|{PossiblePositions T.2} else 1|{PossiblePositions T.2} end
   end
end
   

% Recibe tablero y ficha a usar y devuelve sus posibles posiciones, [nil] si ninungna
fun {IASudominoku Table Tile}
   TableFlatten   = {FlattenUnDetList Table}        % Tablero en una sola lista
   TablePositions = {PossiblePositions TableFlatten} % Posibles posiciones
in
   proc {$ Sol}
      Sudoku Pos1 Pos2 A1 A2
   in
      Sudoku = {FD.list 9*9 1#9}                 % Tablero
      Pos1   = {FD.int 1#9*9}                    % Posición 1 de la ficha
      Pos2   = {FD.int 1#9*9}                    % Posición 2 de la ficha
      A1     = {FD.int 0#1}                      % Variable que indica si la posición 1 está vacía
      A2     = {FD.int 0#1}                      % Variable que indica si la posición 2 está vacía
      {List.forAll {SudokuLists Sudoku} proc {$ E} {FD.distinct E} end}                             % Filas, columnas y cuadros no se repiten
      {List.forAllInd TableFlatten proc {$ I E} if {IsDet E} then {List.nth Sudoku I} =: E end end} % Algunas posiciones ya tienen número
      {FD.distribute ff Sudoku}                  % Tomar cada posibilidad de tablero
      {FD.element Pos1 Sudoku {List.nth Tile 1}} % Posibles posiciones de la ficha
      {FD.element Pos2 Sudoku {List.nth Tile 2}} % Posibles posiciones de la ficha
      {FD.element Pos1 TablePositions A1}        % Estado de posición 1
      {FD.element Pos2 TablePositions A2}        % Estado de posición 2
      A1 =: 1                                    % Debe estar vacía
      A2 =: 1                                    % Debe estar vacía
      % Restricciones para saber que están horizontal o paralelas
      choice
	 % Horizontal(X igual, Y dif en 1)
	 {FD.distance {FD.divI {FD.minus Pos1 1} 9} {FD.divI {FD.minus Pos2 1} 9} '=:' 1}
	 {FD.distance {FD.modI {FD.minus Pos1 1} 9} {FD.modI {FD.minus Pos2 1} 9} '=:' 0}
      []
	 % Vertical(Y igual, X dif. en 1)
	 {FD.distance {FD.divI {FD.minus Pos1 1} 9} {FD.divI {FD.minus Pos2 1} 9} '=:' 0}
	 {FD.distance {FD.modI {FD.minus Pos1 1} 9} {FD.modI {FD.minus Pos2 1} 9} '=:' 1}
      end
      Sol = [Pos1 Pos2]
      {FD.distribute ff Sol}
   end
end

% Retorna las posibles posiciones de todas las fichas disponibles
fun {IAGame Sudoku Tiles}
   case Tiles
   of nil then nil
   [] Ti|Tr then
      local Options in
	 Options = {DropRepeated {SearchOne {IASudominoku Sudoku Ti}} nil} %SearchAll(full IA)
	 % Pasamos las posiciones a coordenadas X,Y del tablero
	 {List.map Options fun {$ E} [(E.1-1) mod 9 + 1 (E.1-1) div 9 + 1 (E.2.1-1) mod 9 + 1 (E.2.1-1) div 9 + 1] end}
	 |{IAGame Sudoku Tr}
      end
   end
end


% Tests
declare
Sudoku0 = [[5 4 8 3 6 1 9 7 2]
	   [1 2 7 4 9 5 3 8 6]
	   [9 6 3 2 8 7 4 5 1]
	   [2 1 4 6 7 8 5 3 9]
	   [7 5 6 9 4 3 2 1 8]
	   [3 8 9 1 5 2 7 6 4]
	   [4 3 1 7 2 6 8 9 5]
	   [6 9 5 8 3 4 1 2 7]
	   [8 7 2 5 1 9 6 4 3]]

Sudoku1 = [[_ _ _ 6 _ _ _ 3 _]
	   [6 _ _ 3 _ _ 4 _ 7]
	   [_ 3 4 _ _ _ _ 8 _]
	   [_ 1 7 _ 2 _ _ _ _]
	   [9 _ 5 7 _ 8 3 _ 1]
	   [_ _ _ _ 4 _ 7 5 _]
	   [_ 4 _ 1 _ _ 8 2 _]
	   [3 _ 1 _ _ 2 _ _ 9]
	   [_ 5 _ _ _ 9 _ _ _]]

Sudoku2 = [[_ _ 3 _ 7 _ _ _ _]
	   [_ _ 1 5 _ _ 8 _ _]
	   [_ 5 _ _ 8 _ _ _ _]
	   [_ 7 _ 8 _ 3 _ 2 _]
	   [4 _ _ _ _ _ _ _ 6]
	   [_ 2 _ 1 _ _ _ 4 _]
	   [_ 3 _ _ 5 _ _ 6 _]
	   [_ _ 2 _ _ 4 9 _ _]
	   [_ _ 9 _ 6 _ 1 _ _]]

Sudoku3 = [[_ _ _ _ _ _ _ _ _]
	   [_ _ _ _ _ _ _ _ _]
	   [_ _ _ _ _ _ _ _ _]
	   [_ 7 _ 9 _ _ _ _ _]
	   [_ 3 _ _ _ _ _ _ 6]
	   [_ _ _ _ _ _ _ 4 _]
	   [_ _ _ _ 2 _ 8 1 _]
	   [_ _ _ _ _ _ 5 _ _]
	   [_ _ _ _ _ _ _ _ _]]

%{Browse {SearchAll {IASudominoku Sudoku1 [1 5]}}}
%{Browse {IAGame Sudoku1 [[1 5] [7 5] [9 1]]}}
%{Browse {IAGame Sudoku3 [[4 7] [5 6] [3 4] [2 3] [6 9] [1 6] [1 8] [7 8] [3 5]]}}