\insert 'functions.oz'
declare [QTk] = {Module.link ["x-oz://system/wp/QTk.ozf"]}

declare CreateCanvas DrawSudoku DrawTiles DrawStats DrawWinner Canvas Width Height in

% Create a canvas of width W and height H to draw in it
proc {CreateCanvas W H}
   local Window TL in
      Width  = W
      Height = H
      TL = td(title:"Sudominoku" canvas(handle:Canvas width:Width height:Height bg:black))
      Window = {QTk.build TL}
      {Window show}
   end
end

% Draw sudoku with numbers specified in matrix Table, LastPos means the last position(domino) to be filled
proc {DrawSudoku Table LastPos}
   % Draw a square and it's current number
   proc {DrawSquare X Y W H Data Color}
      {Canvas create(rect X Y X+W Y+H fill:yellow outline:black)}
      if {IsDet Data} then % If data is defined the draw the square's text
	 {Canvas create(text X+(W div 2) Y+(H div 2) text:{Int.toString Data} fill:Color)}
      end
   end
   Dy = Height div {List.length Table}                         % Height of each square
   Dx = (Width*6 div 10)  div {List.length {List.nth Table 1}} % Width of each square
in
   % Process each matrix element
   for I in 1..{List.length Table} do
      for J in 1..{List.length {List.nth Table I}} do
	 local Color in
	    % If it one of the last positions to be played, color the text red, black otherwise
	    if {Or {And {List.nth LastPos 1} == J {List.nth LastPos 2} == I}
		{And {List.nth LastPos 3} == J {List.nth LastPos 4} == I}} then
	       Color = red
	    else
	       Color = black
	    end
	    {DrawSquare (J-1)*Dx (I-1)*Dy Dx Dy {List.nth {List.nth Table I} J} Color}
	 end
      end
   end
end

% Draw the tiles of domino from each player
proc {DrawTiles TilesCell}
   Tiles = @TilesCell
   fun {TilesText T S}
      case T
      of nil then S
	 % n1|n2  n3|n3 ....
      [] Ti|Tr then {TilesText Tr {List.append S {List.append "  " {List.append {Int.toString {List.nth Ti 1}} {List.append "|" {Int.toString {List.nth Ti 2}}}}}}}
      end
   end
   Dy = (Height*8 div 10) div {List.length Tiles} % Height of each player to draw his tiles
in
   {Canvas create(rect (Width*6 div 10) 0 Width (Height*8 div 10) fill:black)}
   % Process of tiles from all players [ [[n1 n2] [n3 n4]] [[n5 n6] ... ] ... ]
   for I in 1..{List.length Tiles} do
      % Player name in the middle
      {Canvas create(text (Width*8 div 10) (I-1)*Dy+(Dy div 2) text:{TilesText {List.nth Tiles I} {List.append "Player " {List.append {Int.toString I} ":\n"}}} fill:white)}
   end
end

% Draw current player and time left to play
proc {DrawStats Turn TimeLeft}
   {Canvas create(rect (Width*6 div 10) (Height*8 div 10) Width Height fill:gray)} % In grey
   % Turn:Player N  Time Left: 60
   {Canvas create(text (Width*8 div 10) (Height*9 div 10) text:{List.append "STATS:\n" {List.append "Turn: Player " {List.append {Int.toString Turn} {List.append "  Time Left:" {List.append {Int.toString TimeLeft} "s"}}}}} fill:white)}
end

% Draw Winner
proc {DrawWinner Player}
   {Canvas create(rect (Width*6 div 10) (Height*8 div 10) Width Height fill:gray)} % In grey
   % Player N WINS!!!!
   {Canvas create(text (Width*8 div 10) (Height*9 div 10) text:{List.append "Player " {List.append {Int.toString Player} " WINS!!!"}} fill:white)}
end

% Graphical server declaration and offering

declare GraphicalServer Server in
class GraphicalServer
   meth init skip end
   meth createCanvas(width:W height:H) {CreateCanvas W H} end
   meth drawSudoku(table:T lastPos:L) {DrawSudoku T L} end
   meth drawTiles(tiles:T) {DrawTiles T} end
   meth drawStats(player:P time:T) {DrawStats P T} end
   meth drawWinner(player:P) {DrawWinner P} end
end

% Offer the server
{Offer {NewStat GraphicalServer init} "graphserver.txt"}