\insert 'functions.oz'
\insert 'ia.oz'

declare Register DecideGame WaitIA IA Play Turn CTurn Ackn Table Tiles LastPos WaitThreadID IAThreadID in

% Register player with server(judge)
proc {Register}
   proc {TakeTurn T N}
      if N > {List.length T} then {Show "Could not register player"} else
	 if {IsDet {List.nth T N}} then {TakeTurn T N+1} else
	    Turn = N
	    {List.nth T N} = N
	 end
      end
   end
in
   % Create all variables
   WaitThreadID = {NewCell unit}
   IAThreadID   = {NewCell unit}
   {TakeTurn {Take "register.txt"} 1}
   % Retrieve all necessary entities to play on server
   CTurn   = {Take "turn.txt"}    % Turn(1=player1, 2=player2 ...)
   Ackn    = {Take "ackn.txt"}    % Port to tell server: pass or ready
   Table   = {Take "table.txt"}   % Current status of sudoku(matrix with dataflow variables)
   Tiles   = {Take "tiles.txt"}   % List of tiles of all players(cell, where Turn-nth position are the player's tiles)
   LastPos = {Take "lastpos.txt"} % Last position to be played(cell to modify if player made a play)
end

% Decide what tile to play, the one with less options will be played
% Res = Result, Index=Index of winning tale, NOptions=Number of options of current winning tale
fun {DecideGame Options Res Index NOptions} % Res=0 Index=1 NOptions=Sup(Initial state)
   case Options
   of nil then Res
   [] Oi|Or then
      if Oi == nil then {DecideGame Or Res Index+1 NOptions}
      else
	 if {List.length Oi} < NOptions then {DecideGame Or Index Index+1 {List.length Oi}}
	 else {DecideGame Or Res Index+1 NOptions} end
      end
   end
end

% Wait for normal IA
proc {WaitIA Finish}
   WaitThreadID := {Thread.this}  % Update thread ID
   {Delay 60000}                  % Wait Reasonable time
   {Thread.terminate @IAThreadID} % Terminate game IA
   {Send Ackn pass}               % Pass(too much time)
   Finish = unit                  % Mark turn finished
end

% Perform normal IA
proc {IA Finish}
   Options Tile
in
   IAThreadID := {Thread.this}                      % Update thread ID
   Options = {IAGame Table {List.nth @Tiles Turn}}  % Calculate IA
   {Thread.terminate @WaitThreadID}                 % Terminate turn wait
   % Pass or Play
   Tile = {DecideGame Options 0 1 FD.sup}           % Decide tile with less possibilities(if there is)
   if Tile \= 0 then                                % Possible option
      local Option in
	 Option = {List.nth Options Tile}.1
	 LastPos := Option                         % Last Play
	 {MatrixElement Table {List.nth Option 1} {List.nth Option 2}} = {List.nth {List.nth {List.nth @Tiles Turn} Tile} 1} % Table update
	 {MatrixElement Table {List.nth Option 3} {List.nth Option 4}} = {List.nth {List.nth {List.nth @Tiles Turn} Tile} 2} % Table update
      end
      Tiles := {ReplaceElement @Tiles Turn {DropElement {List.nth @Tiles Turn} Tile}}                                        % Tiles update
      {Send Ackn ready}                             % Play the option
   else                                             % No options
      {Send Ackn pass}                              % Pass
   end
   Finish = unit                                    % Mark turn finished
end

% Wait for turn, CurrentTurn is a port where server tells the current player
proc {Play CurrentTurn}
   Finish
in
   case CurrentTurn
   of Ti|Tr then
      if Ti == Turn then % Player's turn
	 thread {WaitIA Finish} end % Wait...
	 thread {IA Finish} end     % ... and think of an option to play
	 {Wait Finish}              % Wait until player made a decision, play or pass
      end
      {Play Tr}                     % Wait for next turn
   end
end

{Register}   % Register itself
{Play CTurn} % Play