\insert 'functions.oz'

declare GUI Table Tiles Players PortTurn Turn PortAckn Ackn LastPos StartGame GameOver LoopGame Winner in

proc {StartGame Rows Columns NPlayers}

   % Create a NxM matrix with unbounded dataflow variables
   fun {CreateTable N M}
      if M == 0 then nil else {List.make N}|{CreateTable N M-1} end
   end

   % Create Tiles(1|2 1|3 1|4 ... 2|3 2|4 ... 8|9)
   fun {CreateTiles N M Max}
      if N == Max then nil else
	 if M > Max then {CreateTiles N+1 N+2 Max} else
	    [N M]|{CreateTiles N M+1 Max}
	 end
      end
   end

   % Scramble the initial tiles of the game
   % TilesLeft      = initial tiles
   % TilesScrambled = tiles delivered to each player in random fashion
   % N              = Player to deliver the next random tile
   fun {ScrambleTiles TilesLeft TilesScrambled N}
      if TilesLeft == nil then TilesScrambled else % No more tiles to scramble
	 if N > NPlayers then {ScrambleTiles TilesLeft TilesScrambled 1} else % Restart with player 1 of tiles are still left
	    local NRand in
	       NRand = {OS.rand} mod {List.length TilesLeft} + 1
	       {ScrambleTiles
		% Take out tile form tiles left
		{DropElement TilesLeft NRand}
		% Deliver random tile to player N
		{ReplaceElement TilesScrambled N {List.append {List.nth TilesScrambled N} [{List.nth TilesLeft NRand}]}}
		% Next player
		N+1}
	    end
	 end
      end
   end

   % Set N numbers in random positions in the sudoku table
   proc {RandomizeTable N}
      JRand = {OS.rand} mod {List.length Table} + 1               % Random Row
      IRand = {OS.rand} mod {List.length {List.nth Table 1}} + 1  % Random Column
   in
      if N == 0 then skip else
	 if {IsDet {List.nth {List.nth Table JRand} IRand}} then {RandomizeTable N} else % if already with number
	    {List.nth {List.nth Table JRand} IRand} = N                                  % if not, set number in position
	    {RandomizeTable N-1}
	 end
      end
   end

   % Register players until positions are full
   proc {RegisterPlayers N}
      Register = {List.make N}        % List with avaiable positions
   in
      {Offer Register "register.txt"} % Offer registration
      {List.forAll Register proc {$ P} {Wait P} end} % Wait for each position to be full
   end
   
in
   {OS.srand 0} % Init randomness
   % Offer game entities
   {Offer Table    "table.txt"}   % Sudoku matrix
   {Offer Tiles    "tiles.txt"}   % Player tiles
   {Offer Turn     "turn.txt"}    % Current Turn
   {Offer PortAckn "ackn.txt"}    % Current player status
   {Offer LastPos  "lastpos.txt"} % Last position to be played
   % Init all variables
   Players    = NPlayers
   PortTurn   = {NewPort Turn}
   PortAckn   = {NewPort Ackn}
   LastPos    = {NewCell [0 0 0 0]}
   Tiles      = {NewCell {CreateTiles 1 2 Rows}}
   % Crate sudoku matrix
   Table      = {CreateTable Rows Columns}
   % Scramble and give player tiles
   Tiles     := {ScrambleTiles @Tiles {CreateNullList Players} Players}
   % Randomize table with 9 numbers
   {RandomizeTable Rows}
   % Host game for players to register
   {RegisterPlayers Players}
   
end

fun {GameOver NPasses LastPlay}
   % Every player passed, the last to play wins
   if NPasses == Players then  Winner = LastPlay end
   % A player has no tiles
   if {List.member 0 {List.map @Tiles List.length}} then {List.forAllInd @Tiles proc {$ I L} if {List.length L} == 0 then Winner = I end end} end
   % Sudoku is full, last to play wins
   if {Not {List.member false {List.map {FlattenUnDetList Table} IsDet}}} then Winner = LastPlay end
   % If any of the above was true, the winner is decided
   {IsDet Winner}
end

% Start being the judge of the game
proc {LoopGame T Ack NPasses LastPlay}
   if T > Players then {LoopGame 1 Ack NPasses LastPlay} else % If loop exceeds the number of players, restart with the first
      {GUI drawSudoku(table:Table lastPos:@LastPos)}          % Draw current sudoku
      {GUI drawTiles(tiles:Tiles)}                            % Draw players' tiles
      {Delay 5000}                                            % 5 second delay to see last play
      if {GameOver NPasses LastPlay} then                     % Check if game is over from last turn play
	 {GUI drawWinner(player:Winner)}                      % Draw winner
      else
	 {GUI drawStats(player:T time:60)}                    % Draw current player stats
	 {Send PortTurn T}                                    % Send players the current player
	 case Ack                                             % Wait for current player move
	 of ready|Ar then
	    {LoopGame T+1 Ar 0 T}                             % Player played, no passes, last to play was him
	 [] pass|Ar then
	    {LoopGame T+1 Ar NPasses+1 LastPlay}              % Player passed, passes increment, last to play remains
	 end
      end
   end
end


% Game Init

GUI = {Take "graphserver.txt"}           % Take rendering server
{GUI createCanvas(width:600 height:300)} % Create gui with rendering server
{StartGame 9 9 4}                        % Host a game with 4 players
{LoopGame 1 Ackn 0 0}                    % Start the game